/*
 * Create a thumbnail from a Zeiss laser scanning microscope image.
 * Reads in an LSM file and writes a 2D, subsampled projection to a
 * JPG file.
 *
 * Cris Luengo, May 2005, September 2006
 */

/*
gcc -o makethumbs makethumbs.c         \
    -I/home/diplib/dip/sunOS/include   \
    -L/home/diplib/dip/sunOS/lib       \
    -ldipio -ldip -lm
*/

#include <stdio.h>
#include <string.h>
#include "diplib.h"
#include "dip_point.h"
#include "dip_manipulation.h"
#include "dip_morphology.h"
#include "dip_binary.h"
#include "dipio.h"
#include "dipio_lsm.h"
#include "dipio_jpeg.h"
#include "dipio_ics.h"

int main (int argc, char *argv[]) {

  DIP_FN_DECLARE("main");
  dip_Resources rg = 0;
  dip_String filename, outname;
  dipio_ImageFileInformation fi;
  dip_Image im, mask, tmp1, tmp2;
  dip_IntegerArray sz, sampling, origin;
  dip_FloatArray fsz;
  dip_BooleanArray ps;
  dip_int ndims, nchans, ii;
  dip_float lower, upper;

  if (argc != 2) {
    fprintf (stderr, "Usage: makethumbs filename[.lsm]\n");
    return -1;
  }

  /* Initialize */
  DIPXJ( dip_Initialise () );
  DIPXJ( dipio_Initialise () );
  DIPXJ( dip_ResourcesNew (&rg, 0) );

  /* Get filename */
  DIPXJ( dip_StringNew (&filename, 0, argv[1], rg) );
  DIPXJ( dipio_FileAddExtension (filename, &filename, "lsm", rg) );

  /* Get the image information from the file */
  DIPXJ( dipio_ImageFileInformationNew (&fi, 0, 0, 0, 0, rg) );
  DIPXJ( dipio_ImageReadLSMInfo (fi, filename) );
  sz = fi->dimensions;
  ndims = sz->size;
  
  /* Find the appropriate subsampling */
  DIPXJ( dip_IntegerArrayNew (&sampling, ndims, 1, rg) );
  /*
  if (sz->array[0]>768 || sz->array[1]>768) {
    sampling->array[0] = sampling->array[1] = 4;
  }
  else if (sz->array[0]>384 || sz->array[1]>384) {
    sampling->array[0] = sampling->array[1] = 2;
  }
  */
  sampling->array[0] = sampling->array[1] = 3;

  /* Read the sub-sampled image */
  DIPXJ( dip_ImageNew (&im, rg) );
  DIPXJ( dipio_ImageReadLSM (im, filename, 0, 0, sampling) );

  /* Is it a color image? */
  nchans = 1;
  if (ndims>=3 && sz->array[ndims-1]<=8) {
    ndims = ndims-1;
    nchans = sz->array[ndims];
  }

  /* Do the projection */
  if (ndims>=3) {
    DIPXJ( dip_BooleanArrayNew (&ps, sz->size, DIP_FALSE, rg) );
    for (ii=2; ii<ndims; ii++) {
      ps->array[ii] = DIP_TRUE;
    }
    DIPXJ( dip_Maximum (im, 0, im, ps) );
    DIPXJ( dip_ChangeDimensions (im, 0) ); /* squeeze */
  }

  DIPXJ( dip_ImageGetDimensions (im, &sz, rg) );
  if ((nchans > 1 && sz->size != 3) || (nchans == 1 && sz->size != 2)) {
    fprintf (stderr, "Assertion failed! The thumbnail image has the wrong dimensionality\n");
    return -1;
  }
  if (nchans > 1 && nchans != sz->array[2]) {
    fprintf (stderr, "Assertion failed! The thumbnail image has the wrong number of channels\n");
    return -1;
  }

  /* Create mask for thumbnail stretching */
  DIPXJ( dip_ImageNew (&mask, rg) );
  DIPXJ( dip_ImageNew (&tmp1, rg) );
  if (nchans > 1) {
    DIPXJ( dip_BooleanArrayNew (&ps, 3, DIP_FALSE, rg) );
    ps->array[2] = DIP_TRUE;
    DIPXJ( dip_Maximum (im, 0, tmp1, ps) );
    DIPXJ( dip_ChangeDimensions (tmp1, 0) ); /* squeeze */
    DIPXJ( dip_RangeThreshold (tmp1, tmp1, 0, 0, 1, 0, DIP_TRUE) );
  }
  else {
    DIPXJ( dip_RangeThreshold (im, tmp1, 0, 0, 1, 0, DIP_TRUE) );
  }
  DIPXJ( dip_FloatArrayNew (&fsz, 2, 3, rg) );
  DIPXJ( dip_Opening (tmp1, mask, 0, 0, fsz, DIP_FLT_SHAPE_RECTANGULAR) );
  DIPXJ( dip_BinaryPropagation (mask, tmp1, mask, 1, 0, DIP_FALSE) );
  DIPXJ( dip_Invert (mask, mask) );
  
  /* Stretching the thumbnail */
  DIPXJ( dip_ConvertDataType (im, im, DIP_DT_SFLOAT) );
  DIPXJ( dip_ImageNew (&tmp2, rg) );
  if (nchans > 1) {
    DIPXJ( dip_IntegerArrayNew (&origin, 3, 0, rg) );
    for (ii=0; ii<nchans; ii++) {
      origin->array[2] = ii;
      DIPXJ( dip_GetSlice (im, tmp1, origin, 0, 1) );
      /* Same as dip_ContrastStretch (tmp1, 0, 5, 95, 255, 0, DIP_CST_LINEAR, 0, 0, 0)
         but with a mask. */
      DIPXJ( dip_Percentile (tmp1, mask, tmp2, 5, 0) );
      DIPXJ( dip_GetFloat (tmp2, &lower, 0) );
      DIPXJ( dip_SubFloat (tmp1, tmp1, lower) );
      DIPXJ( dip_Percentile (tmp1, mask, tmp2, 95, 0) );
      DIPXJ( dip_GetFloat (tmp2, &upper, 0) );
      DIPXJ( dip_MulFloat (tmp1, tmp1, 255/upper) );
      DIPXJ( dip_Clip (tmp1, tmp1, 0, 255, DIP_CLIP_BOTH) );
      DIPXJ( dip_PutSlice (tmp1, im, origin, 0, 1) );
    }
  }
  else {
    /* Same as dip_ContrastStretch (im, 0, 5, 95, 255, 0, DIP_CST_LINEAR, 0, 0, 0)
       but with a mask. */
    DIPXJ( dip_Percentile (im, mask, tmp2, 5, 0) );
    DIPXJ( dip_GetFloat (tmp2, &lower, 0) );
    DIPXJ( dip_SubFloat (im, im, lower) );
    DIPXJ( dip_Percentile (im, mask, tmp2, 95, 0) );
    DIPXJ( dip_GetFloat (tmp2, &upper, 0) );
    DIPXJ( dip_MulFloat (im, im, 255/upper) );
    DIPXJ( dip_Clip (im, im, 0, 255, DIP_CLIP_BOTH) );
  }

  /* Swap 2nd and 3rd channels (to match our specific goals - we've got Sytox in the 3rd channel but want it to be green) */
  if (nchans == 3) {
    origin->array[2] = 1;
    DIPXJ( dip_GetSlice (im, tmp1, origin, 0, 1) );
    origin->array[2] = 2;
    DIPXJ( dip_GetSlice (im, tmp2, origin, 0, 1) );
    DIPXJ( dip_PutSlice (tmp1, im, origin, 0, 1) );
    origin->array[2] = 1;
    DIPXJ( dip_PutSlice (tmp2, im, origin, 0, 1) );
  }

  /* Write the image to another file */
  DIPXJ( dip_StringNew (&outname, filename->size+10, 0, rg) );
  strcpy (outname->string, filename->string);
  outname->string[strlen(outname->string)-4] = 0;
  strcat (outname->string, "-thumb.jpg");
  if (nchans > 1) {
    DIPXJ( dipio_ImageWriteJPEG (im, outname, DIPIO_PHM_RGB, 0, 90) );
  }
  else {
    DIPXJ( dipio_ImageWriteJPEG (im, outname, DIPIO_PHM_GREYVALUE, 0, 90) );
  }

dip_error:

  /* Return all allocated memory to the system */
  DIPXC( dip_ResourcesFree (&rg) );

  /* Let libraries clean up after themselves */
  DIPXC( dipio_Exit () );
  DIPXC( dip_Exit () );

  /* Report errors */
  return dip_ErrorWrite (error, 0, 0, stderr);
}
